#include <compiler.h>

using namespace std;
using namespace LTF;

// 'new' tname ['(' { param } ')'] vname
bool parseFlowNew(Tokenizer *tok, Program *p)
{
	EXPECT(Symbol::Ident, "type name", tok, false);
	
	Type *t = p->findType(tok->currToken()->stringValue());
	if(t == NULL)
	{
		cerr << "Unknown type `" << *tok->currToken()->stringValue() << "` @ " << tok->currToken()->Location() << endl;
		return false;
	}
	
	NewCodeAST *nc = new NewCodeAST(t);
	if(nc == NULL)
	{
		cerr << "Unable to create NewCodeAST(t)" << endl;
		return false;
	}

	int args = 0;
	if(tok->nextToken()->Type == Symbol::OpenBracket)
	{
		do {
			tok->nextToken();
			args++;
			ValueAST *vast = NULL;
			switch(tok->currToken()->Type)
			{
				case Symbol::Int: {
					vast = new IntValueAST(tok->currToken()->intValue(), false);
				} break;
				case Symbol::String: {
					vast = new StringValueAST(tok->currToken()->stringValue());
				} break;
				default:
					cerr << "Unsupported parameter type @ " << tok->currToken()->Location() << endl;
					return false;
			}
			if(!nc->addParam(vast))
			{
				cerr << "Failed to add parameter to NewCodeAST" << endl;
				return false;
			}
			tok->nextToken();
		} while(tok->currToken()->Type == Symbol::Comma);
		
		if(args == 0)
		{
			cerr << "No parameters passed to `" << *t->Name << "` @ " << tok->currToken()->Location() << endl;
			cerr << "Either pass some parameters, or do not use ()" << endl;
			return false;
		}
		
		EXPECT(Symbol::CloseBracket, ")", tok, false);
		tok->nextToken();
	}
	if(args != t->countArguments())
	{
		cerr << "Error using type `" << *t->Name << "`" << endl << "Given: " << args << " arguments, but expecting " << t->countArguments() << endl;
		return false;
	}
	
	EXPECT(Symbol::Ident, "variable name", tok, false);
	VariableAST *v = new VariableAST(tok->currToken()->stringValue(), new TypeArrayAST(t)); 
	if(!p->addVariable(v))
	{
		cerr << "Failed to create new variable named `" << *tok->currToken()->stringValue() << "` @ " << tok->currToken()->Location() << endl;
		cerr << "Variable is probably a duplicate of an earlier defined variable" << endl;
	}
	
	nc->Constructs(v);
	p->addOperation(nc);
	
	tok->nextToken();
	return true;
}

bool parseFlowLink(Tokenizer *tok, Program *p)
{
	EXPECT(Symbol::OpenSquiggle, "{", tok, false);
	tok->nextToken();
	EXPECT(Symbol::Ident, "variable name", tok, false);
	// TODO: lookup variable
	tok->nextToken();
	EXPECT(Symbol::CloseSquiggle, "}", tok, false);
	tok->nextToken();
	EXPECT(Symbol::Ident, "function name", tok, false);
	// TODO: lookup (output) function in variables type
	tok->nextToken();
	
	EXPECT(Symbol::OpenSquiggle, "{", tok, false);
	tok->nextToken();
	EXPECT(Symbol::Ident, "variable name", tok, false);
	// TODO: lookup variable
	tok->nextToken();
	EXPECT(Symbol::CloseSquiggle, "}", tok, false);
	tok->nextToken();
	EXPECT(Symbol::Ident, "function name", tok, false);
	// TODO: lookup (input) function in variables type
	tok->nextToken();

	// TODO: connect output to input
	return true;
}

bool parseFlowUnRef(Tokenizer *tok, Program *p)
{
	EXPECT(Symbol::Ident, "variable name", tok, false);
	// TODO: lookup variable name
	tok->nextToken();
	return true;
}

bool parseFlowStart(Tokenizer *tok, Program *p)
{
	EXPECT(Symbol::OpenSquiggle, "{", tok, false);
	tok->nextToken();
	EXPECT(Symbol::Ident, "variable name", tok, false);
	// TODO: lookup variable
	tok->nextToken();
	EXPECT(Symbol::CloseSquiggle, "}", tok, false);
	tok->nextToken();
	EXPECT(Symbol::Ident, "function name", tok, false);
	// TODO: lookup (thread) function in variables type
	tok->nextToken();
	return true;
}

Program *LTF::parseFlow(FILE *f)
{
	if(f == NULL) return NULL;
	
	// create the tokenizer, and load the symbols we support
	Tokenizer *tok = new Tokenizer(f);
	tok->setSymbolChar('(', Symbol::OpenBracket);
	tok->setSymbolChar(')', Symbol::CloseBracket);
	tok->setSymbolChar(',', Symbol::Comma);
	tok->setSymbolChar('{', Symbol::OpenSquiggle);
	tok->setSymbolChar('}', Symbol::CloseSquiggle);
	tok->setSymbolString("link", Symbol::Link);
	tok->setSymbolString("unref", Symbol::UnRef);
	tok->setSymbolString("new", Symbol::New);
	tok->setSymbolString("start", Symbol::Start);
	tok->setSymbolString("import", Symbol::Import);

	Program *p = new Program();
	
	tok->nextToken();

	while(tok->currToken()->Type != Symbol::EndOfFile) {
		if(tok->currToken()->Type == Symbol::OpenBracket) { // start caring what is here ...
			// eat it
			tok->nextToken();
			switch(tok->currToken()->Type) {
				case Symbol::Import: {
					tok->nextToken();
					if(!parseImport(tok, p))
						return NULL;
				} break;
				case Symbol::New: {
					tok->nextToken();
					if(!parseFlowNew(tok, p))
						return NULL;
				} break;
				case Symbol::Link: {
					tok->nextToken();
					if(!parseFlowLink(tok, p))
						return NULL;
				} break;
				case Symbol::UnRef: {
					tok->nextToken();
					if(!parseFlowUnRef(tok, p))
						return NULL;
				} break;
				case Symbol::Start: {
					tok->nextToken();
					if(!parseFlowStart(tok, p))
						return NULL;
				} break;
				default : {
					std::cerr << tok->currToken()->stringValue() << " Not supported @ " << tok->currToken()->Location() << " In base" << std::endl;
				} break;
			}
			
			EXPECT(Symbol::CloseBracket, ")", tok, NULL);
		}
		tok->nextToken();
	}

	
	return p;
}
