#include <compiler.h>


using namespace LTF;
using namespace std;

static bool parseTypeFunctionArgs(Tokenizer *tok, Program *p, Type *t, FunctionAST *func)
{
	EXPECT(Symbol::OpenBracket, "(", tok, false);
	tok->nextToken();
	while(tok->currToken()->Type != Symbol::CloseBracket)
	{
		Type *nt = NULL;
		switch(tok->currToken()->Type)
		{
			case Symbol::Bits: {
				nt = new TypeBits(false);
			} break;
			case Symbol::SBits: {
				nt = new TypeBits(true);
			} break;
			case Symbol::FBits: {
				nt = new TypeFBits();
			} break;
			default: {
				EXPECT(Symbol::Ident, "type name", tok, false);
				nt = p->findType(tok->currToken()->stringValue());
				if(nt == NULL)
				{
					cerr << "Type '" << *tok->currToken()->stringValue() << "' not found" << endl;
					return false;
				}
			}
		}
		tok->nextToken();
		TypeArrayAST *type = new TypeArrayAST(nt);
	
		while(tok->currToken()->Type == Symbol::OpenSquare)
		{
			tok->nextToken();
			unsigned long long size = 0;
			if(tok->currToken()->Type == Symbol::Int) { // fixed size array
				size = tok->currToken()->intValue();
				if(size == 0) {
					cerr << "Arrays of zero length are treated as variable length, this is probably not what you meant @ " << tok->currToken()->Location() << endl << "If this is indeed what you intended, then delete the number and just have [] " << endl;
					return false;
				}
				tok->nextToken();
			}
			type->addLevel(size);
			EXPECT(Symbol::CloseSquare, "]", tok, false);
			tok->nextToken();
		}
		func->addParam(type);
		EXPECT(Symbol::CloseBracket && tok->currToken()->Type != Symbol::Comma, ") or ,", tok, false);
	}
	EXPECT(Symbol::CloseBracket, ")", tok, false);
	tok->nextToken();
	
	return true;
}

static bool parseTypeInput(Tokenizer *tok, Program *p, Type *t)
{
	tok->nextToken();
	EXPECT(Symbol::Ident, "input function name", tok, false);
	InputFunctionAST *input = new InputFunctionAST(tok->currToken()->stringValue());
	
	t->addFunction(input);
	
	tok->nextToken();
	return parseTypeFunctionArgs(tok, p, t, input);
}

static bool parseTypeOutput(Tokenizer *tok, Program *p, Type *t)
{
	tok->nextToken();
	EXPECT(Symbol::Ident, "output function name", tok, false);
	OutputFunctionAST *output = new OutputFunctionAST(tok->currToken()->stringValue());
	
	t->addFunction(output);
	
	tok->nextToken();
	if(!parseTypeFunctionArgs(tok, p, t, output))
		return false;

	if(tok->currToken()->Type == Symbol::Default)
	{
		output->setDefault();
		tok->nextToken();
	}
	
	if(tok->currToken()->Type == Symbol::Notify)
	{
		output->setNotify();
		tok->nextToken();
	}
	
	EXPECT(Symbol::CloseBracket, ")", tok, false);

	return true;
}

static bool parseTypeFunction(Tokenizer *tok, Program *p, Type *t)
{
	tok->nextToken();
	EXPECT(Symbol::Ident, "function name", tok, false);
	FunctionAST *func = new FunctionAST(tok->currToken()->stringValue());
	
	t->addFunction(func);
	
	tok->nextToken();
	return parseTypeFunctionArgs(tok, p, t, func);
}

static bool parseTypeType(Tokenizer *tok, Program *p, Type *t)
{
	// Type will be first
	Type *nt = NULL;
	switch(tok->currToken()->Type)
	{
		case Symbol::Bits: {
			nt = new TypeBits(false);
		} break;
		case Symbol::SBits: {
			nt = new TypeBits(true);
		} break;
		case Symbol::FBits: {
			nt = new TypeFBits();
		} break;
		default: {
			EXPECT(Symbol::Ident, "type name", tok, false);
			nt = p->findType(tok->currToken()->stringValue());
			if(nt == NULL)
			{
				cerr << "Type '" << *tok->currToken()->stringValue() << "' not found" << endl;
				return false;
			}
		}
	}
	
	// check for
	// (class c inline (c ...
	if(t == nt && t->isInline())
	{
		cerr << "Inlines cannot contain themselves @" << tok->currToken()->stringValue() << endl;
		return false;
	} 

	tok->nextToken();
	
	TypeArrayAST *type = new TypeArrayAST(nt);
	
	while(tok->currToken()->Type == Symbol::OpenSquare)
	{
		tok->nextToken();
		unsigned long long size = 0;
		if(tok->currToken()->Type == Symbol::Int) { // fixed size array
			size = tok->currToken()->intValue();
			if(size == 0) {
				std::cerr << "Arrays of zero length are treated as variable length, this is probably not what you meant @ " << tok->currToken()->Location() << std::endl << "If this is indeed what you intended, then delete the number and just have [] " << std::endl;
				return false;
			}
			tok->nextToken();
		}
		type->addLevel(size);
		EXPECT(Symbol::CloseSquare, "]", tok, false);
		tok->nextToken();
	}
	
	EXPECT(Symbol::Ident, "variable name", tok, false);
	
	TypeVariableAST *nv = new TypeVariableAST(tok->currToken()->stringValue(), type);
	
	t->addMember(nv);

	tok->nextToken();

	if(tok->currToken()->Type == Symbol::Init)
	{
		nv->setInit();
		tok->nextToken();
	}

	return true;
}

static bool parseTypeClass(Tokenizer *tok, Program *p)
{
	// Type definition
	EXPECT(Symbol::Ident, "class name", tok, false);
	Type *t = p->findType(tok->currToken()->stringValue());
	if(t != NULL)
	{
		cerr << "Redefinition of " << *tok->currToken()->stringValue() << " @ " << tok->currToken()->Location() << endl;
		return false;
	}
	t = new Type(tok->currToken()->stringValue());
	p->addType(t);
	
	tok->nextToken();

	// Handle flags
	if(tok->currToken()->Type == Symbol::Inline)
	{
		if(!t->setInline())
		{
			cerr << "Failed setting inline attribute @ " << tok->currToken()->Location() << " on " << *t->Name << endl;
			return false;
		}
		tok->nextToken();
	}else
	if(tok->currToken()->Type == Symbol::NoLock)
	{
		if(!t->setNoLock())
		{
			cerr << "Failed setting nolock attribute @ " << tok->currToken()->Location() << " on " << *t->Name << endl;
			return false;
		}
		tok->nextToken();
	} else
	if(tok->currToken()->Type == Symbol::NoRef)
	{
		if(!t->setNoRef())
		{
			cerr << "Failed setting noref attribute @ " << tok->currToken()->Location() << " on " << *t->Name << endl;
			return false;
		}
		tok->nextToken();
	}
	
	// Handle inherientance
	if(tok->currToken()->Type == Symbol::Of)
	{
		tok->nextToken();
		EXPECT(Symbol::Ident,"parent class name",tok,NULL);
		Type *pt = p->findType(tok->currToken()->stringValue());
		if(pt == NULL)
		{
			cerr << "No such class '" << *tok->currToken()->stringValue() << "' used @ " << tok->currToken()->Location() << endl;
			return false;
		}
		if(!t->setParent(pt))
		{
			cerr << "Failed setting parent class '" << *tok->currToken()->stringValue() << "' @ " << tok->currToken()->Location() << endl;
			return false;
		}
		tok->nextToken();
	}

	if(tok->currToken()->Type == Symbol::CloseBracket)
	{
		cerr << "Class definition without any contents" << endl;
		return false;
	}

	while(tok->currToken()->Type != Symbol::CloseBracket)
	{
		if(tok->currToken()->Type == Symbol::OpenBracket)
		{
			tok->nextToken();
			Symbol::Type_t currType = tok->currToken()->Type;
			switch(currType)
			{
				case Symbol::Input: {
					if(!parseTypeInput(tok, p, t))
						return false;
				} break;
				case Symbol::Output: {
					if(!parseTypeOutput(tok, p, t))
						return false;	
				} break;
				case Symbol::Function: {
					if(!parseTypeFunction(tok, p, t))
						return false;
				} break;
				case Symbol::Bits:
				case Symbol::SBits:
				case Symbol::FBits:
				case Symbol::Ident: {
					if(!parseTypeType(tok, p, t))
						return false;
				} break;
				default:{
					cerr << "Unexpected " << *tok->currToken()->stringValue() << " @ " << tok->currToken()->Location() << endl;
					return false;
				} break;
			}
			
			EXPECT(Symbol::CloseBracket, ")", tok, false);
		}
		tok->nextToken();
	}
	return true;
}

static bool parseType(FILE *f, Program *p)
{
	if(f == NULL || p == NULL) return false;

	Tokenizer *tok = new Tokenizer(f);
	if(tok == NULL) return false;
	
	tok->setSymbolChar('(', Symbol::OpenBracket);
	tok->setSymbolChar(')', Symbol::CloseBracket);
	tok->setSymbolChar(',', Symbol::Comma);
	tok->setSymbolChar('[', Symbol::OpenSquare);
	tok->setSymbolChar(']', Symbol::CloseSquare);
	tok->setSymbolString("new", Symbol::New);
	tok->setSymbolString("thread", Symbol::Thread);
	tok->setSymbolString("import", Symbol::Import);
	tok->setSymbolString("class", Symbol::Class);
	tok->setSymbolString("of", Symbol::Of);
	tok->setSymbolString("nolock", Symbol::NoLock);
	tok->setSymbolString("noref", Symbol::NoRef);
	tok->setSymbolString("inline", Symbol::Inline);
	tok->setSymbolString("init", Symbol::Init);
	tok->setSymbolString("function", Symbol::Function);
	tok->setSymbolString("input", Symbol::Input);
	tok->setSymbolString("output", Symbol::Output);
	tok->setSymbolString("default", Symbol::Default);
	tok->setSymbolString("notify", Symbol::Notify);
	tok->setSymbolString("bits", Symbol::Bits);
	tok->setSymbolString("sbits", Symbol::SBits);
	tok->setSymbolString("fbits", Symbol::FBits);

	tok->nextToken();

	while(tok->currToken()->Type != Symbol::EndOfFile) {
		if(tok->currToken()->Type == Symbol::OpenBracket) { // start caring what is here ...
			// eat it
			tok->nextToken();
			switch(tok->currToken()->Type)
			{
				case Symbol::Import: {
					if(!parseImport(tok, p))
						return false;
				} break;
				case Symbol::Class: {
					tok->nextToken();
					if(!parseTypeClass(tok, p))
						return false;
				} break;
				case Symbol::CloseBracket:
					break;
				default:
					cerr << *tok->currToken()->stringValue() << ": Not supported in base" << endl;
					return false;
			}
			EXPECT(Symbol::CloseBracket, ")", tok, false);
		}
		tok->nextToken();
	}
	
	delete tok;
	return true;
}

bool LTF::parseImport(Tokenizer *tok, Program *p)
{
	// load this file if it hasn't already been loaded...
	EXPECT(Symbol::String,"name of type file", tok, NULL);
	// try opening the file
	// TODO: make this magical to handle search paths
	FILE *ft = fopen(&((*tok->currToken()->stringValue())[0]), "r");
	if(ft == NULL)
	{
		cerr << endl << "Failed opening type file (" << *tok->currToken()->stringValue() << "): " << strerror(errno) << endl;
		return false;
	}
	if(!parseType(ft, p))
	{
		fclose(ft);
		cerr << endl << "Failed importing type file (" << *tok->currToken()->stringValue() << ") @ " << tok->currToken()->Location() << endl;
		return false;
	}
	tok->nextToken();
	return true;
}
